
/*!
\file
\brief Robust
\author Vasek Smidl 

 */

#include "estim/arx.h"
#include "robustlib.h"
#include <vector>
#include <iostream>
#include <fstream>
#include <itpp/itsignal.h>

using namespace itpp;
using namespace bdm;

const int emlig_size = 2;
const int utility_constant = 10;


int main ( int argc, char* argv[] ) {
	
	itpp::Laplace_RNG LapRNG = Laplace_RNG();

	/*
	// EXPERIMENT: 100 AR model generated time series of length of 30 from y_t=0.95*y_(t-1)+0.05*y_(t-2)+0.2*e_t, 
	// where e_t is normally, student(4) and cauchy distributed are tested using robust AR model, to obtain the 
	// variance of location parameter estimators and compare it to the classical setup.
	vector<vector<vector<string>>> string_lists;
	string_lists.push_back(vector<vector<string>>());
	string_lists.push_back(vector<vector<string>>());
	string_lists.push_back(vector<vector<string>>());

	char* file_strings[3] = {"c:\\ar_normal.txt", "c:\\ar_student.txt", "c:\\ar_cauchy.txt"};
	

	for(int i = 0;i<3;i++)
	{	
		ifstream myfile(file_strings[i]);
		if (myfile.is_open())
		{
			while ( myfile.good() )
			{
				string line;
				getline(myfile,line);
				
				vector<string> parsed_line;
				while(line.find(',') != string::npos)
				{
					int loc = line.find(',');
					parsed_line.push_back(line.substr(0,loc));
					line.erase(0,loc+1);					
				}				

				string_lists[i].push_back(parsed_line);
			}
			myfile.close();
		}
	}

	for(int j = 0;j<string_lists.size();j++)
	{ 
		
		for(int i = 0;i<string_lists[j].size()-1;i++)
		{
			vector<vec> conditions;
			//emlig* emliga = new emlig(2);
			RARX* my_rarx = new RARX(2,30);

			for(int k = 1;k<string_lists[j][i].size();k++)
			{
				vec condition;
				//condition.ins(0,1);				
				condition.ins(0,string_lists[j][i][k]);				
				conditions.push_back(condition);

				//cout << "orig:" << condition << endl;

				if(conditions.size()>1)
				{		
					conditions[k-2].ins(0,string_lists[j][i][k]);
					
				}

				if(conditions.size()>2)
				{
					conditions[k-3].ins(0,string_lists[j][i][k]);

					//cout << "modi:" << conditions[k-3] << endl;

					my_rarx->bayes(conditions[k-3]);

					
					//if(k>5)
					//{
					//	cout << "MaxLik coords:" << emliga->minimal_vertex->get_coordinates() << endl;
					//}
					
				}				
				
			}

			//emliga->step_me(0);
			/*
			ofstream myfile;
			myfile.open("c:\\robust_ar1.txt",ios::app);
			myfile << my_rarx->minimal_vertex->get_coordinates()[0] << ";";
			myfile.close();

			myfile.open("c:\\robust_ar2.txt",ios::app);
			myfile << emliga->minimal_vertex->get_coordinates()[1] << ";";
			myfile.close();
			

			cout << "MaxLik coords:" << emliga->minimal_vertex->get_coordinates() << endl;
			cout << "Step: " << i << endl;
		}

		cout << "One experiment finished." << endl;

		ofstream myfile;
		myfile.open("c:\\robust_ar1.txt",ios::app);
		myfile << endl;
		myfile.close();

		myfile.open("c:\\robust_ar2.txt",ios::app);
		myfile << endl;
		myfile.close();
	}*/
    
	
	// EXPERIMENT: A moving window estimation and prediction of RARX is tested on data generated from 
    // y_t=0.95*y_(t-1)+0.05*y_(t-2)+0.2*e_t, where e_t is normally, student(4) and cauchy distributed. It
    // can be compared to the classical setup.
	

	vector<vector<string>> strings;

	char* file_strings[3] = {"c:\\dataCDClosePercDiff","c:\\ar_student_single","c:\\ar_cauchy_single"};

	for(int i = 0;i<3;i++)
	{			
		char dfstring[80];
		strcpy(dfstring,file_strings[i]);
		strcat(dfstring,".txt");
		
		ifstream myfile(dfstring);
		if (myfile.is_open())
		{			
			string line;
			getline(myfile,line);
				
			vector<string> parsed_line;
			while(line.find(',') != string::npos)
			{
				int loc = line.find(',');
				parsed_line.push_back(line.substr(0,loc));
				line.erase(0,loc+1);					
			}				

			strings.push_back(parsed_line);
			
			myfile.close();
		}
	}
	
	for(int j = 0;j<strings.size();j++)
	{ 		
		vector<vec> conditions;
		//emlig* emliga = new emlig(2);
		RARX* my_rarx = new RARX(2,10,false);
		
		
		mat V0 = 0.0001 * eye ( 3 );
		ARX* my_arx = new ARX(0.85);
		my_arx->set_statistics ( 1, V0 ); //nu is default (set to have finite moments)
		my_arx->set_constant ( false );
		my_arx->validate();
		

		for(int k = 1;k<strings[j].size();k++)
		{
			vec condition;
			//condition.ins(0,1);				
			condition.ins(0,strings[j][k]);				
			conditions.push_back(condition);

			//cout << "orig:" << condition << endl;

			if(conditions.size()>1)
			{		
				conditions[k-2].ins(0,strings[j][k]);
					
			}

			if(conditions.size()>2)
			{
				conditions[k-3].ins(0,strings[j][k]);

				// cout << "Condition:" << conditions[k-3] << endl;

				my_rarx->bayes(conditions[k-3]);
				//my_rarx->posterior->step_me(1);
				
				vec cond_vec;
				cond_vec.ins(0,conditions[k-3][0]);
				
				my_arx->bayes(cond_vec,conditions[k-3].right(2));
					
				
				if(k>8)
				{
					//my_rarx->posterior->step_me(0);

					//mat samples = my_rarx->posterior->sample_mat(10);

					pair<vec,mat> imp_samples = my_rarx->posterior->importance_sample(1000);

					//cout << imp_samples.first << endl;
					
					vec sample_prediction;
					for(int t = 0;t<imp_samples.first.size();t++)
					{
						vec lap_sample = conditions[k-3].left(2);
						//lap_sample.ins(lap_sample.size(),1.0);
						
						lap_sample.ins(0,LapRNG());

						sample_prediction.ins(0,lap_sample*imp_samples.second.get_col(t));
					}

					
					vec sample_pow = sample_prediction;
					
					// cout << sample_prediction << endl;
					vec poly_coefs;
					double prediction;
					bool stop_iteration = false;
					int en = 1;
					do
					{
						double poly_coef = imp_samples.first*sample_pow/(imp_samples.first*ones(imp_samples.first.size()));

						if(en==1)
						{
							prediction = poly_coef;
						}

						poly_coef = poly_coef*en*fact(utility_constant-2+en)/fact(utility_constant-2);

						if(abs(poly_coef)>numeric_limits<double>::epsilon())
						{
							sample_pow = elem_mult(sample_pow,sample_prediction);
							poly_coefs.ins(0,pow(-1.0,en+1)*poly_coef);
						}
						else
						{
							stop_iteration = true;
						}
						
						en++;

						if(en>20)
						{
							stop_iteration = true;
						}
					}
					while(!stop_iteration);

					/*
					ofstream myfile_coef;						

					myfile_coef.open("c:\\coefs.txt",ios::app);
					
					for(int t = 0;t<poly_coefs.size();t++)
					{
						myfile_coef << poly_coefs[t] << ",";					
					}

					myfile_coef << endl;
					myfile_coef.close();
					*/

					//cout << "Coefficients: " << poly_coefs << endl;
										
					/*
					vec bas_coef = vec("1.0 2.0 -8.0");
					cout << "Coefs: " << bas_coef << endl;
					cvec actions2 = roots(bas_coef);
					cout << "Roots: " << actions2 << endl;
					*/
					


					cvec actions = roots(poly_coefs);
					

					bool is_max = false;
					for(int t = 0;t<actions.size();t++)
					{
						if(actions[t].imag() == 0)
						{
							double second_derivative = 0;
							for(int q = 1;q<poly_coefs.size();q++)
							{
								second_derivative+=poly_coefs[q]*pow(actions[t].real(),q-1)*q;
							}

							if(second_derivative<0)
							{
								cout << "Action:" << actions[t].real() << endl;

								is_max = true;
							}
						}
					}

					if(!is_max)
					{
						cout << "No maximum." << endl;
					}

					// cout << "MaxLik coords:" << my_rarx->posterior->minimal_vertex->get_coordinates() << endl;

					/*
					double prediction = 0;
					for(int s = 1;s<samples.rows();s++)
					{
						
						double avg_parameter = imp_samples.get_row(s)*ones(samples.cols())/samples.cols();

						prediction += avg_parameter*conditions[k-3][s-1];

						
						
						/*
						ofstream myfile;
						char fstring[80];
						strcpy(fstring,file_strings[j]);

						char es[5];
						strcat(fstring,itoa(s,es,10));

						strcat(fstring,"_res.txt");
						

						myfile.open(fstring,ios::app);
						
						//myfile << my_rarx->posterior->minimal_vertex->get_coordinates()[0];
						myfile << avg_parameter;
						
						if(k!=strings[j].size()-1)
						{
							myfile << ",";
						}
						else
						{
							myfile << endl;
						}
						myfile.close();
						*/

					
					//}

					// cout << "Prediction: "<< prediction << endl;
					
					enorm<ldmat>* pred_mat = my_arx->epredictor(conditions[k-3].left(2));
					double prediction2 = pred_mat->mean()[0];
					

					ofstream myfile;
					char fstring[80];
					char f2string[80];
					strcpy(fstring,file_strings[j]);
					strcpy(f2string,fstring);

					strcat(fstring,"pred.txt");
					strcat(f2string,"2pred.txt");
					

					myfile.open(fstring,ios::app);
					
					// myfile << my_rarx->posterior->minimal_vertex->get_coordinates()[0];
					myfile << prediction;
					
					if(k!=strings[j].size()-1)
					{
						myfile << ",";
					}
					else
					{
						myfile << endl;
					}
					myfile.close();

					
					myfile.open(f2string,ios::app);
					myfile << prediction2;
					
					if(k!=strings[j].size()-1)
					{
						myfile << ",";
					}
					else
					{
						myfile << endl;
					}
					myfile.close();
					//*/

				}					
			}	
			
			//emliga->step_me(0);
			/*
			ofstream myfile;
			myfile.open("c:\\robust_ar1.txt",ios::app);
			myfile << my_rarx->minimal_vertex->get_coordinates()[0] << ";";
			myfile.close();

			myfile.open("c:\\robust_ar2.txt",ios::app);
			myfile << emliga->minimal_vertex->get_coordinates()[1] << ";";
			myfile.close();
			

			cout << "MaxLik coords:" << emliga->minimal_vertex->get_coordinates() << endl;
			cout << "Step: " << i << endl;*/
		}


	}


	// EXPERIMENT: One step ahead price prediction. Comparison of classical and robust model using optimal trading
    //             with maximization of logarithm of one-step ahead wealth.

	
		
		/*
		cout << "One experiment finished." << endl;

		ofstream myfile;
		myfile.open("c:\\robust_ar1.txt",ios::app);
		myfile << endl;
		myfile.close();

		myfile.open("c:\\robust_ar2.txt",ios::app);
		myfile << endl;
		myfile.close();*/
	

	//emlig* emlig1 = new emlig(emlig_size);

	//emlig1->step_me(0);
	//emlig* emlig2 = new emlig(emlig_size);
	
	/*
	emlig1->set_correction_factors(4);

	for(int j = 0;j<emlig1->correction_factors.size();j++)
	{
		for(set<my_ivec>::iterator vec_ref = emlig1->correction_factors[j].begin();vec_ref!=emlig1->correction_factors[j].end();vec_ref++)
		{
			cout << j << "    ";
			
			for(int i=0;i<(*vec_ref).size();i++)
			{
				cout << (*vec_ref)[i];
			}

			cout << endl;
		}
	}*/
	
	/*
    vec condition5 = "1.0 1.0 1.01";//"-0.3 1.7 1.5";

	emlig1->add_condition(condition5);
	//emlig1->step_me(0);


	vec condition1a = "-1.0 1.02 0.5";
	//vec condition1b = "1.0 1.0 1.01";
	emlig1->add_condition(condition1a);
	//emlig2->add_condition(condition1b);

	vec condition2a = "-0.3 1.7 1.5";
	//vec condition2b = "-1.0 1.0 1.0";
	emlig1->add_condition(condition2a);
	//emlig2->add_condition(condition2b);

	vec condition3a = "0.5 -1.01 1.0";
	//vec condition3b = "0.5 -1.01 1.0";

	emlig1->add_condition(condition3a);
	//emlig2->add_condition(condition3b);	

	vec condition4a = "-0.5 -1.0 1.0";
	//vec condition4b = "-0.5 -1.0 1.0";	

	emlig1->add_condition(condition4a);
	//cout << "************************************************" << endl;
	//emlig2->add_condition(condition4b);
	//cout << "************************************************" << endl;
	
	//cout << emlig1->minimal_vertex->get_coordinates();
	
	//emlig1->remove_condition(condition3a);
	//emlig1->step_me(0);
	//emlig1->remove_condition(condition2a);
	//emlig1->remove_condition(condition1a);
	//emlig1->remove_condition(condition5);
	

	//emlig1->step_me(0);
	//emlig2->step_me(0);
	

	// DA SE POUZIT PRO VYPIS DO SOUBORU
	// emlig1->step_me(0);

	//emlig1->remove_condition(condition1);
	
	

	
	
	/*
	for(int i = 0;i<100;i++)
	{
		cout << endl << "Step:" << i << endl;		

		double condition[emlig_size+1];		

		for(int k = 0;k<=emlig_size;k++)
		{
			condition[k] = (rand()-RAND_MAX/2)/1000.0;		
		}
			

		vec* condition_vec = new vec(condition,emlig_size+1);
		emlig1->add_condition(*condition_vec);

		/*
		for(polyhedron* toprow_ref = emlig1->statistic.rows[emlig_size]; toprow_ref != emlig1->statistic.end_poly; toprow_ref = toprow_ref->next_poly)
		{
			cout << ((toprow*)toprow_ref)->probability << endl;
		}
		*/
		/*
		cout << emlig1->statistic_rowsize(emlig_size) << endl << endl;
	
		/*
		if(i-emlig1->number_of_parameters >= 0)
		{
			pause(30);
		}
		*/

		// emlig1->step_me(i);
		
		/*
		vector<int> sizevector;
		for(int s = 0;s<=emlig1->number_of_parameters;s++)
		{
			sizevector.push_back(emlig1->statistic_rowsize(s));
		}
		*/
	//}
    


	
	/*
	emlig1->step_me(1);

	vec condition = "2.0 0.0 1.0";	

	emlig1->add_condition(condition);

	vector<int> sizevector;
	for(int s = 0;s<=emlig1->number_of_parameters;s++)
	{
		sizevector.push_back(emlig1->statistic_rowsize(s));
	}

	emlig1->step_me(2);

	condition = "2.0 1.0 0.0";

	emlig1->add_condition(condition);

	sizevector.clear();
	for(int s = 0;s<=emlig1->number_of_parameters;s++)
	{
		sizevector.push_back(emlig1->statistic_rowsize(s));
	}
	*/

	return 0;
}



